/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         express.cpp
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Aug 9, 1998
    o+
    o+     A high-level view of expressions.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#include <string.h>
#include <assert.h>

#include "express.h"
#include "decl.h"

#include "gram.h"
#include "token.h"

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
/* Print a char, converting chars to escape sequences. */
static
ostream&
MetaChar(ostream& out, char c, bool inString)
{
    switch (c)
    {
      case '\'':
        if (inString) 
          out << "'";
        else
          out << "\\'";
        break;

      case '"':
        if (inString) 
          out << "\\\"";
        else
          out << "\"";
        break;

      case '\0':
        out << "\\0";
        break;
      case '\\':
        out << "\\\\";
        break;
      case '\n':
        out << "\\n";
        break;
      case '\t':
        out << "\\t";
        break;
      case '\r':
        out << "\\r";
        break;
      case '\f':
        out << "\\f";
        break;
      case '\b':
        out << "\\b";
        break;
      case '\v':
        out << "\\v";
        break;
      case '\a':
        out << "\\a";
        break;
      default:
        out << c;
    }

    return out;
}

/*  ###############################################################  */
/* Print a string, converting chars to escape sequences. */
ostream&
MetaString(ostream& out, char *string)
{
    while (*string)
    {
      MetaChar(out,*string,true);
      string++;
    }

    return out;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
static
void
PrintAssignOp(ostream& out, AssignOp op)
{
    switch (op)
    {
      case AO_Equal:          //  =
        out << "=";
        break;

      case AO_PlusEql:        // +=
        out << "+=";
        break;

      case AO_MinusEql:       // -=
        out << "-=";
        break;

      case AO_MultEql:        // *=
        out << "*=";
        break;

      case AO_DivEql:         // /=
        out << "/=";
        break;

      case AO_ModEql:         // %=
        out << "%=";
        break;

      case AO_ShlEql:         // <<=
        out << "<<=";
        break;

      case AO_ShrEql:         // >>=
        out << ">>=";
        break;

      case AO_BitNotEql:      // ~=
        out << "~=";
        break;

      case AO_BitAndEql:      // &=
        out << "&=";
        break;

      case AO_BitXorEql:      // ^=
        out << "^=";
        break;

      case AO_BitOrEql:       // |=
        out << "|=";
        break;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
static
void
PrintRelOp(ostream& out, RelOp op)
{
    switch (op)
    {
      case RO_Equal:          // ==
        out << "==";
        break;

      case RO_NotEqual:       // !=
        out << "!=";
        break;

      case RO_Less:           // < 
        out << "<";
        break;

      case RO_LessEql:        // <=
        out << "<=";
        break;

      case RO_Grtr:           // > 
        out << ">";
        break;

      case RO_GrtrEql:        // >=
        out << ">=";
        break;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
static
void
PrintUnaryOp(ostream& out, UnaryOp op)
{
    switch (op)
    {
      case UO_Plus:           // +
        out << "+";
        break;

      case UO_Minus:          // -
        out << "-";
        break;

      case UO_BitNot:         // ~
        out << "~";
        break;

      case UO_Not:            // !
        out << "!";
        break;

      case UO_PreInc:         // ++x
      case UO_PostInc:        // x++
        out << "++";
        break;

      case UO_PreDec:         // --x
      case UO_PostDec:        // x--
        out << "--";
        break;

      case UO_AddrOf:         // &
        out << "&";
        break;

      case UO_Deref:          // *
        out << "*";
        break;

      case UO_Sizeof:         // sizeof
        out << "sizeof";
        break;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
static
void
PrintBinaryOp(ostream& out, BinaryOp op)
{
    switch (op)
    {
      case BO_Plus:        // +
        out << "+";
        break;

      case BO_Minus:       // -
        out << "-";
        break;

      case BO_Mult:        // *
        out << "*";
        break;

      case BO_Div:         // /
        out << "/";
        break;

      case BO_Mod:         // %
        out << "%";
        break;

      case BO_Shl:         // <<
        out << "<<";
        break;

      case BO_Shr:         // >>
        out << ">>";
        break;

      case BO_BitAnd:      // &
        out << "&";
        break;

      case BO_BitXor:      // ^
        out << "^";
        break;

      case BO_BitOr:       // |
        out << "|";
        break;

      case BO_And:         // &&
        out << "&&";
        break;

      case BO_Or:          // ||
        out << "||";
        break;

      case BO_Comma:       // x,y
        out << ",";
        break;

      case BO_Member:      // x.y
        out << ".";
        break;

      case BO_PtrMember:   // x->y
        out << "->";
        break;

      default:
      //  case BO_Index        // x[y]
      case BO_Assign:      // An AssignExpr
      case BO_Rel:         // A RelExpr
        break;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression::Expression(ExpressionType et)
{
    etype = et;
}

Expression::~Expression()
{
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
Expression::dup() const
{
    Expression *ret = NULL;

    if (this)
    {
        ret = new Expression(etype);
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Expression::print(ostream& out) const
{
    out << __PRETTY_FUNCTION__ << endl;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Constant::Constant(long val)
         : Expression (ET_Constant)
{
    ctype = CT_Int;
    lng = val;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Constant::Constant(char val)
         : Expression (ET_Constant)
{
    ctype = CT_Char;
    ch = val;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Constant::Constant(double val)
         : Expression (ET_Constant)
{
    ctype = CT_Float;
    doub = val;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Constant::Constant(ConstantType ct)
         : Expression (ET_Constant)
{
    ctype = ct;
    lng   = 0;
}
 
// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Constant::~Constant()
{
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
Constant::dup() const
{
    Constant *ret = NULL;

    if (this)
    {
        ret = new Constant(ctype);

        switch (ctype)
        {
            case CT_Char:
                ret->ch = ch;
                break;

            case CT_Int:
                ret->lng = lng;
                break;

            case CT_Float:
                ret->doub = doub;
                break;

            default:
                assert(0);
                break;
        }
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Constant::print(ostream& out) const
{
    switch (ctype)
    {
        case CT_Char:
          out << "'";
          MetaChar(out,ch,false);
          out << "'";
          break;

        case CT_Int:
          out << lng;
          break;

        case CT_Float:
          out << doub;
          break;

        default:
          break;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
EnumConstant::EnumConstant(Symbol *nme) : Constant(CT_Enum)
{
    name = nme;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
EnumConstant::~EnumConstant()
{
    delete name;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
EnumConstant::dup() const
{
    EnumConstant *ret = NULL;

    if (this)
    {
        ret = new EnumConstant(name);
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
EnumConstant::print(ostream& out) const
{
    out << __PRETTY_FUNCTION__ << endl;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
StringConstant::StringConstant(char *str) : Constant(CT_String) 
{
    buff = new char[strlen(str)+1];
    strcpy(buff,str);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
StringConstant::~StringConstant()
{
    delete [] buff;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
int
StringConstant::length() const
{
    return strlen(buff);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
StringConstant::dup() const
{
    StringConstant *ret = NULL;

    if (this)
    {
        ret = new StringConstant(buff);
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
StringConstant::print(ostream& out) const
{
    out << '"';
    MetaString(out,buff);
    out << '"';
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ArrayConstant::ArrayConstant(): Constant(CT_Array)
{
    size   = 0;
    nItems = 0;

    items  = NULL; 
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ArrayConstant::~ArrayConstant()
{
    for (int j=0; j < nItems; j++)
    {
        delete items[j];
    }

    delete [] items;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
ArrayConstant::addElement( Expression *expr )
{
    if (size == nItems)
    {
        if (size == 0)
            size = 4;
        else
            size += size;

        Expression   **oldItems = items;

        items = new Expression* [size];

        for (int j=0; j < nItems; j++)
        {
            items[j] = oldItems[j];
        }

        delete [] oldItems;
    }

    items[nItems] = expr;
    nItems++;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
ArrayConstant::dup() const
{
    ArrayConstant *ret = NULL;

    if (this)
    {
        ret = new ArrayConstant();

        ret->size  = size;
        ret->items = new Expression* [size];

        for (int j=0; j < nItems; j++)
            ret->addElement(items[j]->dup());
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
ArrayConstant::print(ostream& out) const
{
    out <<  "{ ";

    if (nItems > 0)
    {
        items[0]->print(out);

        for (int j=1; j < nItems; j++)
        {
            out << ", ";
            items[j]->print(out);
        }
    }

    out <<  " }";
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Variable::Variable(Symbol *varname)
         : Expression( ET_Variable )
{
    name = varname;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Variable::~Variable()
{
    delete name;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
Variable::dup() const
{
    Variable *ret = NULL;

    if (this)
    {
        ret = new Variable(name->dup());
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Variable::print(ostream& out) const
{
    out << *name;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
FunctionCall::FunctionCall( Expression *func )
             : Expression( ET_FunctionCall )
{
    function = func;

    size = _nArgs = 0;
    args = NULL;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
FunctionCall::~FunctionCall()
{
    delete function;

    for (int j=0; j < _nArgs; j++)
        delete args[j];

    delete [] args;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
FunctionCall::addArg( Expression *arg )
{
    if (size == _nArgs)
    {
        if (size == 0)
            size = 4;
        else
            size += size;

        Expression **oldArgs = args;

        args = new Expression* [size];

        for (int j=0; j < _nArgs; j++)
        {
            args[j] = oldArgs[j];
        }

        delete [] oldArgs;
    }

    args[_nArgs] = arg;
    _nArgs++;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
FunctionCall::addArgs( treenode *node )
{
    if (node)
    {
        switch (node->hdr.type)
        {
            default:
                {
                  Expression *arg = buildExpr(node);
                  addArg(arg);
                }
                break;

            case TN_EXPR_LIST:
                addArgs(node->lnode);
                addArgs(node->rnode);
                break;
        }
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
FunctionCall::dup() const
{
    FunctionCall *ret = NULL;

    if (this)
    {
        ret = new FunctionCall(function->dup());

        ret->size  = size;
        ret->args = new Expression* [size];

        for (int j=0; j < _nArgs; j++)
            ret->addArg(args[j]->dup());
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
FunctionCall::print(ostream& out) const
{
    function->print(out);

    out << "(";

    if (_nArgs > 0)
    {
        args[0]->print(out);
        for (int j=1; j < _nArgs; j++)
        {
            out << ",";
            args[j]->print(out);
        }
    }

    out << ")";
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
UnaryExpr::UnaryExpr( UnaryOp _op, Expression *expr )
          : Expression( ET_UnaryExpr )
{
    uOp = _op;
    _operand = expr;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
UnaryExpr::~UnaryExpr()
{
    delete _operand;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
UnaryExpr::dup() const
{
    UnaryExpr *ret = NULL;

    if (this)
    {
        ret = new UnaryExpr(uOp,_operand->dup());
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
UnaryExpr::print(ostream& out) const
{
    switch (uOp)
    {
      default:
        PrintUnaryOp(out,uOp);

        if (operand()->precedence() < precedence())
            out << "(" << *operand() << ")";
        else
            out << *operand();
        break;

      case UO_PostInc:
      case UO_PostDec:
        if (operand()->precedence() < precedence())
            out << "(" << *operand() << ")";
        else
            out << *operand();

        PrintUnaryOp(out,uOp);
        break;

      case UO_Sizeof:
        PrintUnaryOp(out,uOp);
        out << "(" << *operand() << ")";
        break;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
int
UnaryExpr::precedence() const
{
    switch (uOp)
    {
      case UO_Plus:           // +
      case UO_Minus:          // -
      case UO_BitNot:         // ~
      case UO_Not:            // !

      case UO_PreInc:         // ++x
      case UO_PreDec:         // --x

      case UO_AddrOf:         // &
      case UO_Deref:          // *
      case UO_Sizeof:         // sizeof
        return 15;

      case UO_PostInc:        // x++
      case UO_PostDec:        // x--
        return 16;
    }

    /* Not reached */
    return 16;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
BinaryExpr::BinaryExpr( BinaryOp _op, Expression *lExpr, Expression *rExpr )
           : Expression( ET_BinaryExpr )
{
    bOp = _op;
    _leftExpr  = lExpr;
    _rightExpr = rExpr;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
BinaryExpr::~BinaryExpr()
{
    delete _leftExpr;
    delete _rightExpr;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
BinaryExpr::dup() const
{
    BinaryExpr *ret = NULL;

    if (this)
    {
        ret = new BinaryExpr(bOp,_leftExpr->dup(),_rightExpr->dup());
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
BinaryExpr::print(ostream& out) const
{
    assert(leftExpr());

    if (leftExpr()->precedence() < precedence())
        out << "(" << *leftExpr() << ")";
    else
        out << *leftExpr();

    bool useSpace = !((bOp == BO_Member) || (bOp == BO_PtrMember));

    if (useSpace)
        out << " ";

    PrintBinaryOp(out,bOp);

    if (useSpace)
        out << " ";

    assert(rightExpr());

    if (rightExpr()->precedence() < precedence())
        out << "(" << *rightExpr() << ")";
    else
        out << *rightExpr();
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
int
BinaryExpr::precedence() const
{
    switch (bOp)
    {
      case BO_Plus:        // +
      case BO_Minus:       // -
        return 12;

      case BO_Mult:        // *
      case BO_Div:         // /
      case BO_Mod:         // %
        return 13;

      case BO_Shl:         // <<
      case BO_Shr:         // >>
        return 11;

      case BO_BitAnd:      // &
        return 8;

      case BO_BitXor:      // ^
        return 7;

      case BO_BitOr:       // |
        return 6;

      case BO_And:         // &&
        return 5;

      case BO_Or:          // ||
        return 4;

      case BO_Comma:       // x,y
        return 1;

      case BO_Member:      // x.y
      case BO_PtrMember:   // x->y
        return 16;

      case BO_Assign:      // An AssignExpr
      case BO_Rel:         // A RelExpr
        break;
    }

    return 1;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
TrinaryExpr::TrinaryExpr( Expression *cExpr,
                          Expression *tExpr, Expression *fExpr )
            : Expression( ET_TrinaryExpr )
{
    _condExpr  = cExpr;
    _trueExpr  = tExpr;
    _falseExpr = fExpr;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
TrinaryExpr::~TrinaryExpr()
{
    delete _condExpr;
    delete _trueExpr;
    delete _falseExpr;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
TrinaryExpr::dup() const
{
    TrinaryExpr *ret = NULL;

    if (this)
    {
        ret = new TrinaryExpr(_condExpr->dup(),
                              _trueExpr->dup(),_falseExpr->dup());
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
TrinaryExpr::print(ostream& out) const
{
    out << "(" << *condExpr() << ")";

    out << " ? ";
    out << "(" << *trueExpr() << ")";

    out << " : ";
    out << "(" << *falseExpr() << ")";
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
AssignExpr::AssignExpr( AssignOp _op, Expression *lExpr, Expression *rExpr )
           : BinaryExpr( BO_Assign, lExpr, rExpr )
{
    aOp = _op;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
AssignExpr::~AssignExpr()
{
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
AssignExpr::dup() const
{
    AssignExpr *ret = NULL;

    if (this)
    {
        ret = new AssignExpr(aOp,_leftExpr->dup(),_rightExpr->dup());
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
AssignExpr::print(ostream& out) const
{
    if (lValue()->precedence() < precedence())
        out << "(" << *lValue() << ")";
    else
        out << *lValue();

    out << " ";
    PrintAssignOp(out,aOp);
    out << " ";

    if (rValue()->precedence() < precedence())
        out << "(" << *rValue() << ")";
    else
        out << *rValue();
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
RelExpr::RelExpr( RelOp _op, Expression *lExpr, Expression *rExpr )
        : BinaryExpr( BO_Rel, lExpr, rExpr )
{
    rOp = _op;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
RelExpr::~RelExpr()
{
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
RelExpr::dup() const
{
    RelExpr *ret = NULL;

    if (this)
    {
        ret = new RelExpr(rOp,_leftExpr->dup(),_rightExpr->dup());
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
RelExpr::print(ostream& out) const
{
    if (leftExpr()->precedence() <= precedence())
        out << "(" << *leftExpr() << ")";
    else
        out << *leftExpr();

    out << " ";
    PrintRelOp(out,rOp);
    out << " ";

    if (rightExpr()->precedence() <= precedence())
        out << "(" << *rightExpr() << ")";
    else
        out << *rightExpr();
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
int
RelExpr::precedence() const
{
    switch (rOp)
    {
      case RO_Equal:
      case RO_NotEqual:
        return 9;

      default:
      case RO_Less:
      case RO_LessEql:
      case RO_Grtr:
      case RO_GrtrEql:
        return 10;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
CastExpr::CastExpr(Decl *typeExpr, Expression *operand )
         : Expression( ET_CastExpr )
{
    castTo = typeExpr;
    expr = operand;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
CastExpr::~CastExpr()
{
    delete castTo;
    delete expr;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
CastExpr::dup() const
{
    CastExpr *ret = NULL;

    if (this)
    {
        ret = new CastExpr(castTo->dup(),expr->dup());
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
CastExpr::print(ostream& out) const
{
    out << "(";
    castTo->print(out,true); 
    out << ") ";

    expr->print(out);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
IndexExpr::IndexExpr( Expression *_array, Expression *sub )
          : Expression(ET_IndexExpr)
{
    array     = _array;
    _subscript = sub;
}

IndexExpr::~IndexExpr()
{
    delete array;
    delete _subscript;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
IndexExpr::dup() const
{
    IndexExpr *ret = NULL;

    if (this)
    {
        ret = new IndexExpr(array->dup(),_subscript->dup());
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
IndexExpr::print(ostream& out) const
{
    array->print(out);

    out << "[";
    _subscript->print(out);
    out << "]";
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ostream&
operator<< (ostream& out, const Expression& expr)
{
    expr.print(out);
    return out;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
buildExpr(treenode *node)
{
    leafnode   *leaf;
    Expression *lExpr, *rExpr;
    Expression *result = NULL;

    lExpr = rExpr = NULL;

    if (node)
    {
        switch (node->hdr.which)
        {
            default:
            case NONE_T:
            case FOR_T:
                assert(0);
                break;

            case IF_T:
                {
                assert(node->hdr.type == TN_COND_EXPR);

                if_node *inode = (if_node*) node;
                Expression *cond  = buildExpr(inode->cond);
                Expression *tExpr = buildExpr(inode->then_n);
                Expression *fExpr = buildExpr(inode->else_n);

                result = new TrinaryExpr(cond, tExpr, fExpr);
                }
                break;
    
            case LEAF_T:
                leaf = (leafnode*) node;
                switch (leaf->hdr.tok)
                {
                    case CHAR_CONST:
                        result = new Constant((char)leaf->data.cval);
                        break;
    
                    case INUM:
                        result = new Constant((long)leaf->data.ival);
                        break;
    
                    case RNUM:
                        result = new Constant(leaf->data.dval);
                        break;
    
                    case STRING:
                        result = new StringConstant(leaf->data.str);
                        break;

                    case IDENT:
                        {
                        Symbol *sym = buildSymbol(leaf);
                        result = new Variable(sym);
                        }
                        break;
                }
                break;
    
            case NODE_T:
                switch (node->hdr.type)
                {
                    default:
                        break;

                    case TN_EXPR:
                        lExpr = buildExpr(node->lnode);
                        rExpr = buildExpr(node->rnode);

                        switch (node->hdr.tok)
                        {
                            case PLUS:
                                if (lExpr)
                                  result = new BinaryExpr(BO_Plus,lExpr,rExpr);
                                else
                                  result = new UnaryExpr(UO_Plus,rExpr);
                                break;

                            case MINUS:
                                if (lExpr)
                                  result = new BinaryExpr(BO_Minus,lExpr,rExpr);
                                else
                                  result = new UnaryExpr(UO_Minus,rExpr);
                                break;

                            case STAR:
                                result = new BinaryExpr(BO_Mult,lExpr,rExpr);
                                break;

                            case DIV:
                                result = new BinaryExpr(BO_Div,lExpr,rExpr);
                                break;

                            case MOD:
                                result = new BinaryExpr(BO_Mod,lExpr,rExpr);
                                break;
            
                            case NOT:
                                result = new UnaryExpr(UO_Not,rExpr);
                                break;
            
                            case AND:
                                result = new BinaryExpr(BO_And,lExpr,rExpr);
                                break;

                            case OR:
                                result = new BinaryExpr(BO_Or,lExpr,rExpr);
                                break;
            
                            case B_NOT:
                                result = new UnaryExpr(UO_BitNot,rExpr);
                                break;

                            case B_AND:
                                result = new BinaryExpr(BO_BitAnd,lExpr,rExpr);
                                break;
            
                            case B_OR:
                                result = new BinaryExpr(BO_BitOr,lExpr,rExpr);
                                break;
            
                            case B_XOR:
                                result = new BinaryExpr(BO_BitXor,lExpr,rExpr);
                                break;
            
                            case L_SHIFT:
                                result = new BinaryExpr(BO_Shl,lExpr,rExpr);
                                break;
            
                            case R_SHIFT:
                                result = new BinaryExpr(BO_Shr,lExpr,rExpr);
                                break;

                            case COMMA:
                                result = new BinaryExpr(BO_Comma,lExpr,rExpr);
                                break;
            
                            case INCR:          // ++
                                if (rExpr)
                                  result = new UnaryExpr(UO_PreInc,rExpr);
                                else
                                  result = new UnaryExpr(UO_PostInc,lExpr);
                                break;

                            case DECR:          // --
                                if (rExpr)
                                  result = new UnaryExpr(UO_PreDec,rExpr);
                                else
                                  result = new UnaryExpr(UO_PostDec,lExpr);
                                break;

                            // Relational Ops
                            case EQUAL:         // ==
                                result = new RelExpr(RO_Equal,lExpr,rExpr);
                                break;

                            case NOT_EQ:
                                result = new RelExpr(RO_NotEqual,lExpr,rExpr);
                                break;

                            case LESS:
                                result = new RelExpr(RO_Less,lExpr,rExpr);
                                break;

                            case LESS_EQ:
                                result = new RelExpr(RO_LessEql,lExpr,rExpr);
                                break;

                            case GRTR:
                                result = new RelExpr(RO_Grtr,lExpr,rExpr);
                                break;

                            case GRTR_EQ:
                                result = new RelExpr(RO_GrtrEql,lExpr,rExpr);
                                break;
                        }
                        break;
    
                    case TN_ASSIGN:
                        lExpr = buildExpr(node->lnode);
                        rExpr = buildExpr(node->rnode);

                        switch (node->hdr.tok)
                        {
                            case EQ:        // =
                                result = new AssignExpr(AO_Equal,lExpr,rExpr);
                                break;

                            case PLUS_EQ:
                                result = new AssignExpr(AO_PlusEql,
                                                        lExpr,rExpr);
                                break;

                            case MINUS_EQ:
                                result = new AssignExpr(AO_MinusEql,
                                                        lExpr,rExpr);
                                break;

                            case STAR_EQ:
                                result = new AssignExpr(AO_MultEql,
                                                        lExpr,rExpr);
                                break;

                            case DIV_EQ:
                                result = new AssignExpr(AO_DivEql,lExpr,rExpr);
                                break;

                            case MOD_EQ:
                                result = new AssignExpr(AO_ModEql,lExpr,rExpr);
                                break;
            
                            case B_NOT_EQ:
                                result = new AssignExpr(AO_BitNotEql,
                                                        lExpr,rExpr);
                                break;
            
                            case B_AND_EQ:
                                result = new AssignExpr(AO_BitAndEql,
                                                        lExpr,rExpr);
                                break;
            
                            case B_OR_EQ:
                                result = new AssignExpr(AO_BitOrEql,
                                                        lExpr,rExpr);
                                break;
            
                            case B_XOR_EQ:
                                result = new AssignExpr(AO_BitXorEql,
                                                        lExpr,rExpr);
                                break;
            
                            case L_SHIFT_EQ:
                                result = new AssignExpr(AO_ShlEql,lExpr,rExpr);
                                break;
            
                            case R_SHIFT_EQ:
                                result = new AssignExpr(AO_ShrEql,lExpr,rExpr);
                                break;
                        }
                        break;
    
                    case TN_CAST:
                        {
                        Decl *decl = new Decl();
                        buildDecl(decl,node->lnode);
                        rExpr = buildExpr(node->rnode);

                        result = new CastExpr(decl,rExpr);
                        }
                        break;

                    case TN_DEREF:
                        rExpr = buildExpr(node->rnode);
                        result = new UnaryExpr(UO_Deref,rExpr);
                        break;

                    case TN_ADDROF:
                        rExpr = buildExpr(node->rnode);
                        result = new UnaryExpr(UO_AddrOf,rExpr);
                        break;

                    case TN_INDEX:
                        result = buildIndexExpr(node);
                        break;

                    case TN_EXPR_LIST:
                        lExpr = buildExpr(node->lnode);
                        rExpr = buildExpr(node->rnode);
                        result = new BinaryExpr(BO_Comma,lExpr,rExpr);
                        break;

                    case TN_SELECT:
                        lExpr = buildExpr(node->lnode);
                        rExpr = buildExpr(node->rnode);

                        switch (node->hdr.tok)
                        {
                          case DOT:
                            result = new BinaryExpr(BO_Member,lExpr,rExpr);
                            break;

                          case ARROW:
                            result = new BinaryExpr(BO_PtrMember,lExpr,rExpr);
                            break;

                          default:
                            break;
                        }
                        break;

                    case TN_FUNC_CALL:
                        {
                        lExpr = buildExpr(node->lnode);
                        FunctionCall *fCall = new FunctionCall(lExpr);
                        fCall->addArgs(node->rnode);
                        result = fCall;
                        }
                        break;

                    case TN_INIT_BLK:
                    case TN_INIT_LIST:
                        result = buildArrayInit(NULL,node);
                        break;
                }
                break;
        }
    }

    return result;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ArrayConstant*
buildArrayInit(ArrayConstant *array, treenode *node)
{
    if (node)
    {
        switch (node->hdr.type)
        {
            default:
                {
                  Expression *expr = buildExpr(node);
                  array->addElement(expr);
                }
                break;
    
            case TN_INIT_BLK:
                {
                  ArrayConstant *ac = new ArrayConstant();

                  buildArrayInit(ac,node->lnode);
                  buildArrayInit(ac,node->rnode);

                  if (array)
                      array->addElement(ac);
                  else
                      array = ac;
                }
                break;
    
            case TN_INIT_LIST:
                buildArrayInit(array,node->lnode);
                buildArrayInit(array,node->rnode);
                break;
        }
    }

    return array;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
buildIndexExpr(treenode *node)
{
    // NYI - expand to handle multidimensional arrays directly.

    Expression *array = buildExpr(node->lnode);
    Expression *sub   = buildExpr(node->rnode);

    return new IndexExpr(array,sub);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
#define    SHOW(X)    case X: return #X
char*
nameOfExpressionType( ExpressionType type )
{
    switch (type)
    {
        default:
            return "Unknown ExpressionType";

        SHOW(ET_VoidExpr);

        SHOW(ET_Constant);
        SHOW(ET_Variable);
        SHOW(ET_FunctionCall);

        SHOW(ET_AssignExpr);
        SHOW(ET_RelExpr);

        SHOW(ET_UnaryExpr);
        SHOW(ET_BinaryExpr);
        SHOW(ET_TrinaryExpr);

        SHOW(ET_CastExpr);
        SHOW(ET_IndexExpr);
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
char*
nameOfConstantType(ConstantType type)
{
    switch (type)
    {
        default:
            return "Unknown ConstantType";

        SHOW(CT_Char);
        SHOW(CT_Int);
        SHOW(CT_Float);

        SHOW(CT_Enum);
        SHOW(CT_String);
        SHOW(CT_Array);
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
char*
nameOfAssignOp(AssignOp op)
{
    switch (op)
    {
        default:
            return "Unknown AssignOp";

        SHOW(AO_Equal);          //  =

        SHOW(AO_PlusEql);        // +=
        SHOW(AO_MinusEql);       // -=
        SHOW(AO_MultEql);        // *=
        SHOW(AO_DivEql);         // /=
        SHOW(AO_ModEql);         // %=

        SHOW(AO_ShlEql);         // <<=
        SHOW(AO_ShrEql);         // >>=

        SHOW(AO_BitNotEql);      // ~=
        SHOW(AO_BitAndEql);      // &=
        SHOW(AO_BitXorEql);      // ^=
        SHOW(AO_BitOrEql);       // |=
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
char*
nameOfRelOp(RelOp op)
{
    switch (op)
    {
        default:
            return "Unknown RelOp";

        SHOW(RO_Equal);          // ==
        SHOW(RO_NotEqual);       // !=

        SHOW(RO_Less);           // < 
        SHOW(RO_LessEql);        // <=
        SHOW(RO_Grtr);           // > 
        SHOW(RO_GrtrEql);        // >=
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
char*
nameOfUnaryOp(UnaryOp op)
{
    switch (op)
    {
        default:
            return "Unknown UnaryOp";

        SHOW(UO_Plus);           // +
        SHOW(UO_Minus);          // -
        SHOW(UO_BitNot);         // ~
        SHOW(UO_Not);            // !

        SHOW(UO_PreInc);         // ++x
        SHOW(UO_PreDec);         // --x
        SHOW(UO_PostInc);        // x++
        SHOW(UO_PostDec);        // x--

        SHOW(UO_AddrOf);         // &
        SHOW(UO_Deref);          // *

        SHOW(UO_Sizeof);         // sizeof
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
char*
nameOfBinaryOp(BinaryOp op)
{
    switch (op)
    {
        default:
            return "Unknown BinaryOp";

        SHOW(BO_Plus);        // +
        SHOW(BO_Minus);       // -
        SHOW(BO_Mult);        // *
        SHOW(BO_Div);         // /
        SHOW(BO_Mod);         // %

        SHOW(BO_Shl);         // <<
        SHOW(BO_Shr);         // >>
        SHOW(BO_BitAnd);      // &
        SHOW(BO_BitXor);      // ^
        SHOW(BO_BitOr);       // |

        SHOW(BO_And);         // &&
        SHOW(BO_Or);          // ||

        SHOW(BO_Comma);       // x,y

        SHOW(BO_Member);      // x.y
        SHOW(BO_PtrMember);   // x->y

        SHOW(BO_Assign);      // An AssignExpr
        SHOW(BO_Rel);         // A RelExpr
    }
}

#undef SHOW
// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
