/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         express.h
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Aug 9, 1998
    o+
    o+     A high-level view of expressions.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#ifndef    EXPRESS_H
#define    EXPRESS_H

#include <stdlib.h>
#include <iostream.h>

#include "symbol.h"

#include "tree.h"


class Decl;
class Type;

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum ExpressionType
{
    ET_VoidExpr=0,           // No value

    ET_Constant,
    ET_Variable,
    ET_FunctionCall,

    ET_AssignExpr,
    ET_RelExpr,

    ET_UnaryExpr,
    ET_BinaryExpr,
    ET_TrinaryExpr,          // x ? y : z

    ET_CastExpr,             // (type) x
    ET_IndexExpr             // x[y]...[]
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum ConstantType
{
    CT_Char,
    CT_Int,
    CT_Float,

    CT_Enum,
    CT_String,
    CT_Array            // For an array initializer
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum AssignOp
{
    AO_Equal,          //  =

    AO_PlusEql,        // +=
    AO_MinusEql,       // -=
    AO_MultEql,        // *=
    AO_DivEql,         // /=
    AO_ModEql,         // %=

    AO_ShlEql,         // <<=
    AO_ShrEql,         // >>=

    AO_BitNotEql,      // ~=
    AO_BitAndEql,      // &=
    AO_BitXorEql,      // ^=
    AO_BitOrEql        // |=
};

enum RelOp
{
    RO_Equal,          // ==
    RO_NotEqual,       // !=

    RO_Less,           // < 
    RO_LessEql,        // <=
    RO_Grtr,           // > 
    RO_GrtrEql         // >=
};

enum UnaryOp
{
    UO_Plus=0,         // +
    UO_Minus,          // -
    UO_BitNot,         // ~
    UO_Not,            // !

    UO_PreInc,         // ++x
    UO_PreDec,         // --x
    UO_PostInc,        // x++
    UO_PostDec,        // x--

    UO_AddrOf,         // &
    UO_Deref,          // *

    UO_Sizeof          // sizeof
};

enum BinaryOp
{
    BO_Plus,        // +
    BO_Minus,       // -
    BO_Mult,        // *
    BO_Div,         // /
    BO_Mod,         // %

    BO_Shl,         // <<
    BO_Shr,         // >>
    BO_BitAnd,      // &
    BO_BitXor,      // ^
    BO_BitOr,       // |

    BO_And,         // &&
    BO_Or,          // ||

    BO_Comma,       // x,y

    BO_Member,      // x.y
    BO_PtrMember,   // x->y

    BO_Assign,      // An AssignExpr
    BO_Rel          // A RelExpr
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Expression
{
  public:
    Expression(ExpressionType et);
    virtual ~Expression();

    virtual int precedence() const { return 16; }

    virtual Expression *dup() const;

    virtual void print(ostream& out) const;

    ExpressionType    etype;

  friend ostream& operator<< ( ostream& out, const Expression& expr );
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Constant : public Expression
{
  public:
    Constant(long val);
    Constant(char val);
    Constant(float val);
    Constant(double val);
    Constant(ConstantType ct);
    virtual ~Constant();
 
    virtual Expression *dup() const;
    virtual void print(ostream& out) const;

    ConstantType    ctype;

    union
    {
      long    lng;
      char    ch;
      double  doub;
    };
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class EnumConstant : public Constant
{
  public:
    EnumConstant( Symbol *nme );
   ~EnumConstant();

    Expression *dup() const;
    void print(ostream& out) const;

    Symbol    *name;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class StringConstant : public Constant
{
  public:
    StringConstant(char *str);
   ~StringConstant();

    int length() const;

    Expression *dup() const;
    void print(ostream& out) const;

    char    *buff;    // dynamic string storage
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class ArrayConstant : public Constant
{
  public:
    ArrayConstant();
   ~ArrayConstant();

    void addElement( Expression *expr );

    Expression *dup() const;
    void print(ostream& out) const;

    int           size;
    int           nItems;
    Expression  **items;        // Array of whatever, including another array.
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Variable : public Expression
{
  public:
    Variable(Symbol *varname);
   ~Variable();

    Expression *dup() const;
    void print(ostream& out) const;

    Symbol    *name;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class FunctionCall : public Expression
{
  public:
    FunctionCall( Expression *func );
   ~FunctionCall();
    
    int  nArgs() const { return _nArgs; }
    void addArg( Expression *arg );

    void addArgs( treenode *args );

    Expression *dup() const;
    void print(ostream& out) const;

    Expression   *function;
    int           size;
    int           _nArgs;
    Expression  **args;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class UnaryExpr : public Expression
{
  public:
    UnaryExpr( UnaryOp op, Expression *expr );
   ~UnaryExpr();

    int precedence() const;

    Expression *operand() const { return _operand; }

    UnaryOp op() const { return uOp; }

    Expression *dup() const;
    void print(ostream& out) const;

    UnaryOp    uOp;
    Expression *_operand;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class BinaryExpr : public Expression
{
  public:
    BinaryExpr( BinaryOp op, Expression *lExpr, Expression *rExpr );
   ~BinaryExpr();

    Expression *leftExpr() const { return _leftExpr; }
    Expression *rightExpr() const { return _rightExpr; }

    virtual int precedence() const;

    BinaryOp op() const { return bOp; }

    virtual Expression *dup() const;
    virtual void print(ostream& out) const;

    BinaryOp    bOp;

    Expression *_leftExpr;
    Expression *_rightExpr;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class TrinaryExpr : public Expression
{
  public:
    TrinaryExpr( Expression *cExpr,
                 Expression *tExpr, Expression *fExpr );
   ~TrinaryExpr();

    Expression *condExpr() const { return _condExpr; }
    Expression *trueExpr() const { return _trueExpr; }
    Expression *falseExpr() const { return _falseExpr; }

    int precedence() const { return 3; }

    Expression *dup() const;
    void print(ostream& out) const;

    Expression *_condExpr;
    Expression *_trueExpr;
    Expression *_falseExpr;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class AssignExpr : public BinaryExpr
{
  public:
    AssignExpr( AssignOp op, Expression *lExpr, Expression *rExpr );
   ~AssignExpr();

    Expression *lValue() const { return leftExpr(); }
    Expression *rValue() const { return rightExpr(); }

    AssignOp op() const { return aOp; }

    int precedence() const { return 2; }

    Expression *dup() const;
    void print(ostream& out) const;

    AssignOp    aOp;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class RelExpr : public BinaryExpr
{
  public:
    RelExpr( RelOp op, Expression *lExpr, Expression *rExpr );
   ~RelExpr();

    int precedence() const;

    RelOp op() const { return rOp; }

    Expression *dup() const;
    void print(ostream& out) const;

    RelOp    rOp;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class CastExpr : public Expression
{
  public:
    CastExpr( Decl *typeExpr, Expression *operand );
   ~CastExpr();

    Decl *type() { return castTo; }

    int precedence() const { return 14; }

    Expression *dup() const;
    void print(ostream& out) const;

    Decl        *castTo;
    Expression  *expr;    // The expression being cast.
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class IndexExpr : public Expression
{
  // I should extend this to be able to handle
  // multi-dimensional arrays directly.

  public:
    IndexExpr( Expression *_array, Expression *sub );
   ~IndexExpr();
    
    //addSubscript( Expression *sub );
    Expression *subscript(int i);

    Expression *dup() const;
    void print(ostream& out) const;

    Expression   *array;
    Expression   *_subscript;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

Expression    *buildExpr(treenode *node);
Expression    *buildIndexExpr(treenode *node);

ArrayConstant *buildArrayInit(ArrayConstant *array, treenode *node);

// For debugging, show the enumerated name.
char *nameOfExpressionType(ExpressionType type);
char *nameOfConstantType(ConstantType type);
char *nameOfAssignOp(AssignOp op);
char *nameOfRelOp(RelOp op);
char *nameOfUnaryOp(UnaryOp op);
char *nameOfBinaryOp(BinaryOp op);

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

#endif  /* EXPRESS_H */

